#version 450
#extension GL_EXT_buffer_reference : require

#include "hex_common.glsl"

#define PI 3.1415926535897932384626433832795
const float w = 0.5;
const float x = 0.75;
const float z = sqrt(3.0)/2.0;

const vec4 vertices[] = {
  vec4(0,             0, 0,             1),
  vec4(w*cos(0*PI/3), 0, w*sin(0*PI/3), 1),
  vec4(w*cos(1*PI/3), 0, w*sin(1*PI/3), 1),
  vec4(w*cos(2*PI/3), 0, w*sin(2*PI/3), 1),
  vec4(w*cos(3*PI/3), 0, w*sin(3*PI/3), 1),
  vec4(w*cos(4*PI/3), 0, w*sin(4*PI/3), 1),
  vec4(w*cos(5*PI/3), 0, w*sin(5*PI/3), 1),
};

const uint indices[] = {
  0, 2, 1,
  0, 3, 2,
  0, 4, 3,
  0, 5, 4,
  0, 6, 5,
  0, 1, 6,
};

const vec4 starts[] = {
  vec4( 0, 0,    z, 0), //  0, -1
  vec4(-x, 0,  z/2, 0), // -1,  0
  vec4(-x, 0, -z/2, 0), // -1, +1
  vec4( 0, 0,   -z, 0), //  0, +1
  vec4( x, 0, -z/2, 0), // +1,  0
  vec4( x, 0,  z/2, 0), // +1, -1
};

const vec2 start_coords[] = {
  vec2(0, -1),
  vec2(-1, 0),
  vec2(-1, 1),
  vec2(0, 1),
  vec2(1, 0),
  vec2(1, -1),
};

const vec4 direction[] = {
  vec4(-x, 0, -z/2, 0),
  vec4( 0, 0,   -z, 0),
  vec4( x, 0, -z/2, 0),
  vec4( x, 0,  z/2, 0),
  vec4( 0, 0,    z, 0),
  vec4(-x, 0,  z/2, 0),
};

const vec2 direction_coords[] = {
  vec2(-1, 1),
  vec2(0, 1),
  vec2(1, 0),
  vec2(1, -1),
  vec2(0, -1),
  vec2(-1, 0),
};

layout(location=0) out vec4 color;

vec4 int2color(uint color_int) {
    float r = float((color_int >> 24) & 0xFF);
    float g = float((color_int >> 16) & 0xFF);
    float b = float((color_int >>  8) & 0xFF);
    float a = float((color_int >>  0) & 0xFF);
    return vec4(r, g, b, a)/255.0;
}

const uint region_size = 7;
const float region_width = x*(2*region_size-1);
const float region_height = z*(2*region_size-1);
const uint region_hex_count = 3*region_size*(region_size-1)+1;

void main() {
  uint hex_index = gl_InstanceIndex - ((gl_InstanceIndex/region_hex_count)*region_hex_count);
  Region region = pc.context.regions[gl_InstanceIndex/region_hex_count];

  vec2 region_qr = vec2(region.q, region.r);

  vec4 region_pos = vec4(0, 0, 0, 0);
  region_pos.x = (region_qr.x+region_qr.y/2)*region_width - region_qr.y*x/2;
  region_pos.z = 0.75*region_qr.y*region_height + 0.25*region_qr.y*z + 0.5*region_qr.x*z;

  uint radius = 0;
  uint ring = 0;
  uint side = 0;
  uint hex = 0;
  if(hex_index != 0) {
    radius = uint(floor(0.5 + sqrt(12*hex_index-3)/6));
    ring = hex_index - (3*radius*radius - 3*radius + 1);
    side = uint(floor(ring/(radius)));
    hex = ring - (radius*side);
  }
  
  vec4 position = vertices[indices[gl_VertexIndex]] + (starts[side]*radius) + (direction[side]*hex) + region_pos;
  if(gl_VertexIndex % 3 == 0) {
    position.y = (region.hexes[hex_index].heights[0] +
                  region.hexes[hex_index].heights[1] +
                  region.hexes[hex_index].heights[2] +
                  region.hexes[hex_index].heights[3] +
                  region.hexes[hex_index].heights[4] +
                  region.hexes[hex_index].heights[5])/6;
  } else {
    position.y = region.hexes[hex_index].heights[indices[gl_VertexIndex]-1];
  }
  color = int2color(region.hexes[hex_index].colors[indices[gl_VertexIndex]]);

  vec2 hex_qr = start_coords[side]*radius + direction_coords[side]*hex;

  /*
  vec4 hex_pos = vec4(0, 0, 0, 1);
  hex_pos.x = x*hex_qr.x;
  hex_pos.z = -z*hex_qr.y - z*hex_qr.x/2;

  vec4 world_pos = vec4(0, 0, 0, 1);
  world_pos.x = hex_pos.x + region_pos.x;
  world_pos.z = hex_pos.z + region_pos.z;

  vec2 world_qr = vec2(0, 0);
  world_qr.x = world_pos.x/x;
  world_qr.y = -world_pos.z/z - world_pos.x/(2*x);
  */

  vec2 world_qr;
  world_qr.x = (x*hex_qr.x + (region_qr.x+region_qr.y/2)*region_width - region_qr.y*x/2)/x;
  world_qr.y = -(-z*hex_qr.y - z*hex_qr.x/2 + 0.75*region_qr.y*region_height + 0.25*region_qr.y*z + 0.5*region_qr.x*z)/z - (x*hex_qr.x + (region_qr.x+region_qr.y/2)*region_width - region_qr.y*x/2)/(2*x);

  color = vec4(world_qr/region_size, 0.0, 1.0);

  gl_Position = pc.context.proj * pc.context.view * position;
}
